unit DabEmployee;

interface

uses
  Classes, CustomVisitor;

type
  TPayStrategy = class;

  TDabEmployee = class(TCollectionItem)
  private
    FFirstName   : string;
    FSurname     : string;
    FPosition    : string;
    FPayStrategy : TPayStrategy;

    function GetFullName  : string;
    function GetPayAmount : Double;
  protected
  public
    destructor Destroy; override;

    procedure Assign(Source : TPersistent); override;
    // Needed for Visitor
    procedure Accept(Visitor : TCustomVisitor); virtual;
  published
    property FirstName : string read FFirstName write FFirstName;
    property Surname   : string read FSurname write FSurname;
    property Position  : string read FPosition write FPosition;
    property FullName  : string read GetFullName;
    // This is the context interface for the Strategy pattern
    property PayAmount : Double read GetPayAmount;
  end;


  TDabEmployees = class(TOwnedCollection)
  private
  protected
    function  GetItem(Index : Integer) : TDabEmployee;
    procedure SetItem(Index : Integer;Value : TDabEmployee);
  public
    // GoF suggest a constructor that returns the iterator that
    // matches this list type. If you had more than one type of iterator for
    // each list type then this approach requires an extra parameter and
    // makes it hard to attach create more iterators
    // In any case Delphi doesn't allow constructors like this :
    //
    //        constructor CreateIterator : TObject;
    //
    // Use a factory singleton instead

    function  Add : TDabEmployee;
    procedure AddNew(const AFirstName,ASurname,APosition : string;
                     APayStrategy : TPayStrategy);
    // Needed for Visitor
    procedure Accept(Visitor : TCustomVisitor); virtual;
    function  FindItemID(ID : Integer) : TDabEmployee;
    function  Insert(Index : Integer) : TDabEmployee;

    property Items[Index : Integer] : TDabEmployee read GetItem
                                                   write SetItem; default;
  published
  end;


  TPayStrategy = class(TObject)
  private
    // The employee is the GoF "context", although in this case more information
    // is needed (see descendants), and the employee is not actually used
    FEmployee : TDabEmployee;
  protected
    // Function doesn't have to be abstract. It could return a default value.
    function GetPayAmount : Double; virtual; abstract;

    property Employee : TDabEmployee read FEmployee write FEmployee;
  public
    property PayAmount : Double read GetPayAmount;
  end;


implementation

uses SysUtils;

{ TDabEmployee }

destructor TDabEmployee.Destroy;
begin
  FPayStrategy.Free;
  inherited;
end;


function TDabEmployee.GetFullName : string;
begin
  Result := TrimLeft(FirstName + ' ') + Surname;
end;


function TDabEmployee.GetPayAmount : Double;
begin
  // Default behaviour
  Result := 0;

  if Assigned(FPayStrategy) then begin
    Result := FPayStrategy.PayAmount;
  end;
end;


procedure TDabEmployee.Assign(Source : TPersistent);
begin
  if Source is TDabEmployee then begin
    FirstName := TDabEmployee(Source).FirstName;
    Surname   := TDabEmployee(Source).Surname;
    Position  := TDabEmployee(Source).Position;
  end;
end;


procedure TDabEmployee.Accept(Visitor : TCustomVisitor);
begin
  Visitor.Visit(Self);
end;


{ TDabEmployees }

function TDabEmployees.GetItem(Index : Integer) : TDabEmployee;
begin
  Result := TDabEmployee(inherited GetItem(Index));
end;


procedure TDabEmployees.SetItem(Index : Integer;Value : TDabEmployee);
begin
  inherited SetItem(Index,Value);
end;


function TDabEmployees.Add : TDabEmployee;
begin
  Result := TDabEmployee(inherited Add);
end;


procedure TDabEmployees.AddNew(const AFirstName,ASurname,APosition : string;
                               APayStrategy : TPayStrategy);
begin
  with Add do begin
    FirstName    := AFirstName;
    Surname      := ASurname;
    Position     := APosition;
    FPayStrategy := APayStrategy;
  end;
end;


procedure TDabEmployees.Accept(Visitor : TCustomVisitor);
  var
    i : Integer;
begin
  Visitor.Visit(Self);

  for i := 0 to Count - 1 do begin
    Items[i].Accept(Visitor);
  end;
end;


function TDabEmployees.FindItemID(ID : Integer) : TDabEmployee;
begin
  Result := TDabEmployee(inherited FindItemID(ID));
end;


function TDabEmployees.Insert(Index : Integer) : TDabEmployee;
begin
  Result := TDabEmployee(inherited Insert(Index));
end;


end.
 